import numpy as np
import unittest
import slam.texture as stex
import slam.io as sio
import slam.sulcal_depth as sdepth

acceptable_error = 0.0001

class TestSulcalDepth(unittest.TestCase):
    test_mesh = sio.load_mesh('examples/data/example_mesh.gii')
    test_dpf = sio.load_texture('examples/data/example_dpf.gii').darray[0]
    test_dpf_star = sio.load_texture('examples/data/example_dpf_star.gii').darray[0]
    dpf = sdepth.depth_potential_function(test_mesh)[0]
    dpf_star = sdepth.dpf_star(test_mesh)[0]
    def test_basic(self):
        # test size
        self.assertTrue(self.test_dpf.shape==self.dpf.shape)
        self.assertTrue(self.test_dpf_star.shape == self.dpf_star.shape)

    def test_consistency(self):
        self.assertTrue(np.max(np.abs(self.test_dpf-self.dpf)) < acceptable_error)
        self.assertTrue(np.max(np.abs(self.test_dpf_star - self.dpf_star)) < acceptable_error)

if __name__ == "__main__":

    unittest.main()



import os
import numpy as np
from slam import (io, differential_geometry,
                  geodesics, texture,
                  curvature, vertex_voronoi,
                  sulcal_depth, utils)


def compute_mesh_features(mesh, save=True, outdir=None, check_if_exist=True):
    """
    Function that computes the mean curvature, the depth potential function
    and the voronoi areas of a mesh.
    Returns numpy arrays.
    Parameters
    ----------
    mesh
    save
    outdir
    check_if_exist

    Returns
    -------

    """

    if not outdir:
        outdir = ''

    path_to_mean_curvature = os.path.join(outdir, "mean_curvature.gii")
    path_to_dpf = os.path.join(outdir, "dpf.gii")
    path_to_voronoi = os.path.join(outdir, "voronoi.npy")

    # Get vertex mean curvature
    if check_if_exist and os.path.exists(path_to_mean_curvature):
        mean_curvature = io.load_texture(path_to_mean_curvature).darray[0]
    else:
        print("\n\tComputing the curvature\n")
        PrincipalCurvatures, _, _ = curvature.curvatures_and_derivatives(mesh)
        mean_curv = (
                0.5 * (PrincipalCurvatures[0, :] + PrincipalCurvatures[1, :]))
        filt_mean_curv = \
            (utils.z_score_filtering(np.array([mean_curv]), z_thresh=3))
        mean_curvature = filt_mean_curv[0]

    # Get vertex depth potential function
    if check_if_exist and os.path.exists(path_to_dpf):
        dpf = io.load_texture(path_to_dpf).darray
    else:
        print("\n\tComputing the DPF\n")
        # returns a list of dpf for different alpha values
        dpf = sulcal_depth.dpf_star(mesh,
                                    alphas=[0.03],
                                    curvature=mean_curvature)
        dpf = dpf[0]

    # Get vertex voronoi areas
    if check_if_exist and os.path.exists(path_to_voronoi):
        voronoi = np.load(path_to_voronoi)
    else:
        print("\n\tComputing Voronoi's vertex\n")
        voronoi = vertex_voronoi.vertex_voronoi(mesh)

    if save:
        curv_tex = texture.TextureND(darray=mean_curvature)
        io.write_texture(curv_tex, path_to_mean_curvature)
        dpf_tex = texture.TextureND(darray=dpf)
        io.write_texture(dpf_tex, path_to_dpf)
        np.save(path_to_voronoi, voronoi)

    return mean_curvature, dpf, voronoi


def normalize_thresholds(mesh, voronoi, thresh_dist=20.0,
                         thresh_ridge=1.5, thresh_area=50.0, side="left"):
    """
    Function that normalizes the thresholds for the watershed algorithm.
    Threshold on distance between pits is normalized by the Fiedler geodesic
    length.
    Threshold on basin's area is normalized by the surface area of the mesh.

    :args: mesh(trimesh object): surface triangular mesh loaded
    with io.load_mesh()
    :args: voronoi(numpy array): voronoi areas of the mesh
    :args: thresh_dist(float): distance threshold for the watershed algorithm
    :args: thresh_ridge(float): ridge threshold for the watershed algorithm
    :args: thresh_area(float): area threshold for the watershed algorithm
    :args: side: side of the brain (left or right)
    :return: normalized thresholds (thresh_dist, thresh_ridge, thresh_area)

    Parameters
    ----------
    mesh
    voronoi
    thresh_dist
    thresh_ridge
    thresh_area
    side

    Returns
    -------

    """

    # Compute Fiedler length and surface area for watershed threshold
    # normalization
    print("\n\tComputing the Fiedler geodesic length and surface area\n")
    fielder = differential_geometry.mesh_laplacian_eigenvectors(mesh, 1)
    imin = fielder.argmin()
    imax = fielder.argmax()
    min_mesh_fiedler_length = geodesics.compute_gdist(mesh, imin, imax)
    mesh_area = np.sum(voronoi)

    # Set group average values
    group_average_fiedler_length = 238.25 if side == "right" else 235.95
    group_average_surface_area = 91433.68 if side == "right" else 91369.33

    # Normalize the thresholds regarding brain size
    thresh_dist *= min_mesh_fiedler_length / group_average_fiedler_length
    thresh_area *= mesh_area / group_average_surface_area

    print("Fielder length", min_mesh_fiedler_length)
    print("Distance threshold :", thresh_dist)

    return thresh_dist, thresh_ridge, thresh_area


def watershed(mesh, voronoi, dpf, thresh_dist, thresh_ridge,
              thresh_area, mask=None):
    """
    Sulcal segmentation and sulcal pits extraction using watershed by
    flooding.

    Reference paper is:
    G. Auzias, L.Brun, C. Deruelle, O. Coulon. 2015. Deep sulcal
    landmarks: algorithmic and conceptual improvements in the definition
     and extraction of sulcal pits. NeuroImage

    A related approach can be found in:
    Rettmann ME, Han X, Xu C, Prince JL. 2002. Automated
    sulcal segmentation using watersheds on the cortical
    surface. Neuroimage. 15:329-344.

    Parameters
    ----------
    mesh : white matter triangular mesh of subject (trimesh object)
    voronoi : voronoi area for each vertex (numpy array)
    dpf : depth measure for each vertex (numpy array)
    thresh_dist : threshold on the distance between pits (unit: mm)
    thresh_ridge : threshold on the ridge height (unit: mm)
    thresh_area : threshold on the basin area (unit: mm²)
    mask : binary array for cingular pole exclusion (with ones in the
    region to exclude)

    Returns
    -------
    basins : dictionary with basin properties
        basins[label] = {}
            'pit_index': vertex index of the sulcal pit
            'pit_depth': depth of the sulcal pit
            'basin_vertices': list of vertices in the basin
            'basin_area': area of the basin
    ridges : dict with properties of the ridge between basin i and basin j
        ridges[(i,j)] = {}
            'ridge_index': vertex index of the ridge point
            'ridge_depth': depth of the ridge point
            'ridge_length': number of vertices along the ridge
    adjacency : adjacency matrix of the basins
        adjacency[i,j] = 1 if basin i and j are adjacent, 0 otherwise
    """

    print('Computing watershed by flooding...')
    print(f'Distance between 2 pits: '
          f'{thresh_dist} mm - Ridge height: {thresh_ridge} mm')

    # Initialize output tables and dictionaries
    n_vertices = mesh.vertices.shape[0]  # vertices coordinates (n, 3) float
    vert_idx = np.arange(n_vertices)  # vertex index
    vert_depth = dpf.reshape(n_vertices)
    vert_area = voronoi.reshape(n_vertices)
    # vertex labels (initialized to -1)
    vert_label = np.full(n_vertices, -1, dtype=np.int64)
    vert_neigh = mesh.vertex_neighbors  # list of lists of neighbors
    basins = {}
    adjacency = np.zeros((1, 1), dtype=np.int64)

    # Apply exclusion mask
    # All nodes included in the exclusion mask are not taken into
    # account in the watershed process
    if mask is None:
        mask = np.zeros(dpf.shape)
    mask_indices = np.where(mask == 1)[0]
    nodes = np.vstack((vert_idx, vert_depth)).T
    nodes = np.delete(nodes, mask_indices, axis=0)

    # Sorting step
    # All nodes of the mesh are sorted by their depth
    # (deepest nodes = highest values first)
    nodes_sorted_by_depth = nodes[nodes[:, 1].argsort()[::-1]]

    # Initialize with first pit: deepest node
    idx = int(nodes_sorted_by_depth[0, 0])
    new_label = 0
    vert_label[idx] = new_label
    basins[0] = {}
    basins[0]['pit_index'] = idx
    basins[0]['basin_vertices'] = [idx]

    for node in nodes_sorted_by_depth[1:]:

        idx = int(node[0])
        # indices of neighbors
        neigh_idx = np.array(vert_neigh[idx]).astype(int)
        # labels of neighbors
        neigh_labels = vert_label[neigh_idx]
        # Keep only labelled neighbors
        neigh_idx = neigh_idx[:, None][neigh_labels != -1]
        neigh_labels = neigh_labels[neigh_labels != -1]
        NL = np.unique(neigh_labels)

        #############################################################
        # Case 1: all of its neighbors are unlabeled. Then, this node
        # corresponds to the deepest point of a new catchment basin.
        #############################################################
        if len(NL) == 0:
            new_label += 1  # max(basins.keys()) + 1
            # Update vertex
            vert_label[idx] = new_label
            # Update basins
            basins[new_label] = {}
            basins[new_label]['pit_index'] = idx
            basins[new_label]['basin_vertices'] = [idx]
            # Update adjacency matrix: increment size by 1
            adjacency = np.pad(adjacency,
                               ((0, 1), (0, 1)),
                               mode='constant',
                               constant_values=0)

        #############################################################
        # Case 2: the node is the neighbor of only one catchment basin.
        # Then, this node is assigned to the corresponding basin.
        #############################################################
        elif len(NL) == 1:
            # Update vertex
            vert_label[idx] = NL[0]
            # Update basin
            basins[NL[0]]['basin_vertices'].append(idx)

        ##############################################################
        # Case 3: the node is the neighbor of two or more catchment basins.
        # Then, this node is a ridge point where each pair of basins join.
        # It is assigned to the basin represented by the deepest neighbor
        # vertex, or the lowest label if same depth.
        # Then the conditions for merging two basins are tested (distance
        # between the two pits and ridge height).
        ################################################################
        else:
            # deepest neighbor
            idx_max_depth = np.argmax(vert_depth[neigh_idx])
            lab = np.min(neigh_labels[idx_max_depth])  # lowest label
            # Update vertex
            vert_label[idx] = lab
            # Update basin
            basins[lab]['basin_vertices'].append(idx)

            # MERGING between pairs of neighbor catchment basins
            # The merging condition is questioned between the intersection
            # label (lab) and all other neighbors.
            # The merging condition is only questioned if the basins
            # have never met yet (no existing ridge point)

            # NL.sort()  # NL is already sorted when using np.unique()
            NL = np.delete(NL, NL == lab)
            for neighbor_lab in NL:
                # Check that basins have not already been merged in the loop
                if (neighbor_lab in np.unique(vert_label)
                        and lab in np.unique(vert_label)):
                    # Forcing label_i < label_j so that
                    # pit(i) deeper than pit(j)
                    label_i = min(lab, neighbor_lab)
                    label_j = max(lab, neighbor_lab)
                    if adjacency[label_i, label_j] == 0:  # first meet

                        merging = 0

                        # Compute ridge height
                        ridge_height = (
                            abs(vert_depth[basins[label_j]['pit_index']]
                                - node[1]))
                        if ridge_height < thresh_ridge:

                            # Compute distance between pits
                            v = geodesics.compute_gdist(
                                mesh,
                                basins[label_i]['pit_index'],
                                basins[label_j]['pit_index'])
                            if v < thresh_dist:
                                # print('merging of',label_j,'into',label_i)
                                merging = 1
                                # Update vertex
                                vert_label[
                                    np.where(vert_label == label_j)[0]]\
                                    = label_i
                                # Update basin
                                [basins[label_i]['basin_vertices'].append(b)
                                 for b in basins[label_j]['basin_vertices']]
                                del basins[label_j]
                                # Update ridges
                                # Add adjacent basins of j to i
                                # and clean out adjacency of j
                                adjacency[label_i, :] = (
                                        adjacency[label_i, :] |
                                        adjacency[label_j, :])
                                adjacency[:, label_i] = (
                                        adjacency[:, label_i] |
                                        adjacency[:, label_j])
                                adjacency[label_j, :] = 0
                                adjacency[:, label_j] = 0

                                if label_j == lab:
                                    # lab disappears and is replaced by the
                                    # parent label which becomes adjacent
                                    # to other neighbors.
                                    # Continue the loop to test merge
                                    # with this new lab.
                                    lab = neighbor_lab

                        # Create a new entry in the ridge dictionary
                        if not merging:
                            adjacency[label_i, label_j] = 1
                            adjacency[label_j, label_i] = 1

    print('Number of basins found:', len(basins))

    #################################################################
    # FILTERING OF SMALL BASINS
    # Last merging step on area criterion after the watershed process
    # Each basin which area is less than thresh_area is merged into the
    # neighbor basin it shares the largest border with.
    ##################################################################

    # Select basins with area < thresh_area
    basins2merge = []
    for label in basins:
        basins[label]['pit_depth'] = vert_depth[basins[label]['pit_index']]
        basins[label]['basin_area'] = (
            np.sum(vert_area[basins[label]['basin_vertices']]))
        if basins[label]['basin_area'] <= thresh_area:
            basins2merge.append([label, basins[label]['basin_area']])
    basins2merge = np.array(basins2merge)

    print('nb of basins to remove:', basins2merge.shape[0])
    if basins2merge.shape[0] > 0:

        print('Filtering of small basins...')

        # Sort by area
        basins2merge = basins2merge[basins2merge[:, 1].argsort()]

        # Filtering
        basin_index = 0
        while basin_index != len(basins2merge):
            basin, area = basins2merge[basin_index]
            basin = int(basin)

            # Neighbor basins are those sharing a ridge
            # point with current basin
            neigh_basins = np.where(adjacency[basin] != 0)[0]
            ridges_length = []
            for nb in neigh_basins:
                # List all ridge points
                ridges_vertices = []
                for v in basins[basin]['basin_vertices']:
                    neighbors_vertices = np.array(vert_neigh[v])
                    [ridges_vertices.append(v)
                     for v in
                     neighbors_vertices[vert_label[neighbors_vertices] == nb]]
                # Get ridge length with each neighbor
                ridges_length.append(len(ridges_vertices))
            # Parent basin: the one sharing the largest border
            parent_basin = np.min(neigh_basins[np.argmax(ridges_length)])

            # print('merging of', basin, 'into', int(parent_basin))

            # Update vertex
            vert_label[np.where(vert_label == basin)[0]] = parent_basin
            # Update basin
            [basins[parent_basin]['basin_vertices'].append(b)
             for b in basins[basin]['basin_vertices']]
            basins[parent_basin]['basin_area'] += area
            del basins[basin]
            # Update ridges matrix
            adjacency[parent_basin, :] = (
                    adjacency[parent_basin, :] |
                    adjacency[basin, :])
            adjacency[:, parent_basin] = (
                    adjacency[:, parent_basin] |
                    adjacency[:, basin])
            adjacency[basin, :] = 0
            adjacency[:, basin] = 0

            # Update basins2merge
            if parent_basin in basins2merge[:, 0]:
                # If new area of parent basin is now above the threshold,
                # remove it from the list of basin to merge
                if basins[parent_basin]['basin_area'] > thresh_area:
                    basins2merge = (
                        np.delete(basins2merge,
                                  np.where(basins2merge[:, 0]
                                           == parent_basin)[0],
                                  axis=0))
                # Else, update its area and sort the table again
                else:
                    basins2merge[
                        np.where(basins2merge[:, 0] == parent_basin)[0], 1]\
                        = basins[parent_basin]['basin_area']
                    basins2merge = basins2merge[basins2merge[:, 1].argsort()]

            basin_index += 1

        print('Number of basins after filtering:', len(basins))

    # Create a dictionary with ridge properties
    ridges = {}
    adjacency_index = np.array(np.where(adjacency != 0)).T
    for i, j in adjacency_index:
        # list all ridge points between basin i and basin j
        ridges_vertices = []
        for v in basins[i]['basin_vertices']:
            neighbors_vertices = np.array(vert_neigh[v])
            [ridges_vertices.append(v)
             for v in neighbors_vertices[vert_label[neighbors_vertices] == j]]
        ridges[(i, j)] = {}
        ridges[(i, j)]['ridge_index'] = (
            ridges_vertices)[np.argmax(vert_depth[ridges_vertices])]
        ridges[(i, j)]['ridge_depth'] = np.max(vert_depth[ridges_vertices])
        ridges[(i, j)]['ridge_length'] = len(ridges_vertices)

    return basins, ridges, adjacency


def get_textures_from_dict(mesh, basins, ridges, save=True, outdir=None):
    """
        Function that returns the textures from the dictionaries outputs of
    the watershed
    Parameters
    ----------
    mesh
    basins
    ridges
    save
    outdir

    Returns
    -------

    """

    if save and not outdir:
        outdir = ''

    vert = np.array(mesh.vertices)
    labels = np.full(vert.shape[0], -1, dtype=np.int64)
    atex_pits = np.zeros(vert.shape[0])
    atex_ridges = np.zeros(vert.shape[0])

    for b in basins.keys():
        labels[basins[b]['basin_vertices']] = b
        atex_pits[basins[b]['pit_index']] = 1

    for i, j in ridges:
        atex_ridges[ridges[(i, j)]['ridge_index']] = 1

    # texture of labels
    tex_labels = texture.TextureND(darray=labels.flatten())
    if save:
        io.write_texture(tex_labels, os.path.join(outdir, "labels.gii"))

    # texture of pits
    tex_pits = texture.TextureND(darray=atex_pits.flatten())
    if save:
        io.write_texture(tex_pits, os.path.join(outdir, "pits_tex.gii"))

    # texture of ridges
    tex_ridges = texture.TextureND(darray=atex_ridges.flatten())
    if save:
        io.write_texture(tex_ridges, os.path.join(outdir, "rigdes_tex.gii"))

    return tex_labels, tex_pits, tex_ridges



import unittest
import numpy as np
import networkx as nx
from slam import sulcal_graph
from slam import texture


class TestSulcalGraph(unittest.TestCase):

    def setUp(self):
        # Create a synthetic graph
        self.graph = nx.Graph()
        self.graph.add_node(0, basin_vertices=[0, 1], pit_index=0, basin_label=0)
        self.graph.add_node(1, basin_vertices=[2, 3], pit_index=2, basin_label=1)
        self.graph.add_edge(0, 1, ridge_index=4)

        # Create a synthetic mesh
        class MockMesh:
            vertices = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0.5, 0.5, 0]])

        self.mesh = MockMesh()

        # Create a synthetic texture
        self.texture = np.array([0.1, 0.2, 0.3, 0.4, 0.5])

    def test_add_node_attribute(self):
        a = np.array([0.0, 1.0, 2.0, 3.0, 4.0])

        g = sulcal_graph.add_node_attribute_to_graph(self.graph, a, name='test', save=False)

        self.assertIn("test", g.nodes[0].keys())
        self.assertEqual(a[0], g.nodes[0]['test'])
        self.assertTrue(
            (a[list(nx.get_node_attributes(g, 'pit_index'))] == list(nx.get_node_attributes(g, 'test'))).all())

    def test_add_edge_attribute(self):
        a = np.array([0.0, 1.0, 2.0, 3.0, 4.0])

        g = sulcal_graph.add_edge_attribute_to_graph(self.graph, a, name='test', save=False)

        self.assertIn("test", g.edges[list(g.edges)[0]].keys())
        self.assertEqual(a[4], g.edges[list(g.edges)[0]]['test'])
        self.assertTrue(
            (a[list(nx.get_edge_attributes(g, 'ridge_index'))] == list(nx.get_edge_attributes(g, 'test'))).all())

    def test_add_mean_value_to_graph(self):
        g = sulcal_graph.add_mean_value_to_graph(self.graph, self.texture, name="mean_texture", save=False)

        self.assertIn("mean_texture", g.nodes[0].keys())
        self.assertAlmostEqual(g.nodes[0]["mean_texture"], 0.15)
        self.assertAlmostEqual(g.nodes[1]["mean_texture"], 0.35)

    def test_get_textures_from_graph(self):
        # Test de la fonction get_textures_from_graph
        tex_labels, tex_pits, tex_ridges = sulcal_graph.get_textures_from_graph(self.graph, self.mesh, save=False)

        self.assertIsInstance(tex_labels, texture.TextureND)
        self.assertIsInstance(tex_pits, texture.TextureND)
        self.assertIsInstance(tex_ridges, texture.TextureND)
        self.assertEqual(tex_labels.darray.shape[1], len(self.mesh.vertices))
        self.assertEqual(tex_pits.darray.shape[1], len(self.mesh.vertices))
        self.assertEqual(tex_ridges.darray.shape[1], len(self.mesh.vertices))

    def test_add_geodesic_distances(self):
        def mock_compute_gdist(mesh, ridge):
            return np.array([0.0, 1.0, 2.0, 3.0, 4.0])

        sulcal_graph.geodesics.compute_gdist = mock_compute_gdist
        g = sulcal_graph.add_geodesic_distances_to_graph(self.graph, self.mesh, save=False)

        self.assertIn("geodesic_distance_btw_ridge_pit_i", g.edges[list(g.edges)[0]].keys())
        self.assertIn("geodesic_distance_btw_ridge_pit_j", g.edges[list(g.edges)[0]].keys())
        self.assertIn("geodesic_distance_btw_pits", g.edges[list(g.edges)[0]].keys())


if __name__ == '__main__':
    unittest.main()


import numpy as np
import unittest
from slam import texture

class TestTexture(unittest.TestCase):
    def test_update_darray(self):
        test_texture = texture.TextureND()
        darray = np.zeros((4, 1))
        test_texture.update_darray(darray)
        assert test_texture.shape == darray.shape

if __name__ == "__main__":

    unittest.main()



import numpy as np
from scipy.sparse.linalg import lgmres
import slam.differential_geometry as sdg
import slam.curvature as scurv
import slam.utils as sutl

########################
# error tolerance for lgmres solver
solver_tolerance = 1e-6
########################


def depth_potential_function(mesh, alphas=None, curvature=None):
    """
    compute the depth potential function of a mesh as desribed in
    Boucher, M., Whitesides, S., & Evans, A. (2009).
    Depth potential function for folding pattern representation,
    registration and analysis.
    Medical Image Analysis, 13(2), 203–14.
    doi:10.1016/j.media.2008.09.001
    :param mesh:
    :param curvature:
    :param alphas:
    :return:
    """
    if alphas is None:
        alphas = [0.3]
    if curvature is None:
        # Comptue mean curvature from principal curvatures
        PrincipalCurvatures, PrincipalDir1, PrincipalDir2 \
            = scurv.curvatures_and_derivatives(mesh)
        mean_curv = (
                0.5 * (PrincipalCurvatures[0, :] + PrincipalCurvatures[1, :]))
        filt_mean_curv = (
            sutl.z_score_filtering(np.array([mean_curv]), z_thresh=3))
        curvature = filt_mean_curv[0]

    L, LB = sdg.compute_mesh_laplacian(mesh, lap_type="fem")
    B = (
        2
        * LB
        * (curvature - (np.sum(curvature * LB.diagonal())
                        / np.sum(LB.diagonal())))
    )
    # be careful with factor 2 used in eq (13)

    dpf = []
    for ind, alpha in enumerate(alphas):
        M = alpha * LB + L / 2
        dpf_t, info = lgmres(M.tocsr(), B, rtol=solver_tolerance)
        dpf.append(dpf_t)

    return dpf


def dpf_star(mesh, alphas=None, adaptation=None, curvature=None):
    """
    compute the depth potential function of a mesh. The scale of
    interest is adapted to the size of the mesh.
    :param mesh: TRIMESH MESH
    :param curvature: TEXTURE (darray)
    :param alphas: LIST : adapt the size of interest for computing
    sulcal depth
    :return: TEXTURE (darray)

    Parameters
    ----------
    adaptation
    """
    if alphas is None:
        alphas = [500]
    if curvature is None:
        # Comptue mean curvature from principal curvatures
        PrincipalCurvatures, PrincipalDir1, PrincipalDir2 \
            = scurv.curvatures_and_derivatives(mesh)
        mean_curv = (
                0.5 * (PrincipalCurvatures[0, :] + PrincipalCurvatures[1, :]))
        filt_mean_curv = (
            sutl.z_score_filtering(np.array([mean_curv]), z_thresh=3))
        curvature = filt_mean_curv[0]
    if adaptation is None:
        adaptation = 'volume_hull'
    # print(source.shape)
    # print(mesh.vertices.shape)
    # adaptation of the scale of interest
    if adaptation == 'volume_hull':
        hull = mesh.convex_hull
        vol_hull = hull.volume
        lc = np.power(vol_hull, 1/3)
    if adaptation == 'volume':
        vol = mesh.volume
        lc = np.power(vol, 1/3)
    if adaptation == 'surface':
        surface = mesh.area
        lc = np.power(surface, 1/2)

    # compute the laplacian
    L, LB = sdg.compute_mesh_laplacian(mesh)

    # dedimensialisation of the laplacian and the curvature
    L = L * np.square(lc)
    curvature = curvature * lc

    # compute the dpf
    B = (
        LB
        * (curvature - (np.sum(curvature * LB.diagonal())
                        / np.sum(LB.diagonal()))))

    dpf = []
    for ind, alpha in enumerate(alphas):
        M = (alpha * LB) + L
        dpf_t, info = lgmres(M.tocsr(), B, rtol=solver_tolerance)
        dpf.append(dpf_t)

    return dpf



import os
import pickle
import numpy as np
from slam import io, geodesics, texture
import networkx as nx
import slam.watershed as swat


def extract_sulcal_graph(side, path_to_mesh, path_to_features,
                         path_to_output, path_to_mask=None):
    """
    Main Function that extracts the sulcal graph from a mesh and saves
    it in the given directory.
    Parameters
    ----------
    side
    path_to_mesh
    path_to_features
    path_to_output
    path_to_mask

    Returns
    -------

    """

    mesh = io.load_mesh(path_to_mesh)

    _, dpf, voronoi = swat.compute_mesh_features(mesh,
                                                 save=True,
                                                 outdir=path_to_features,
                                                 check_if_exist=True)
    thresh_dist, thresh_ridge, thresh_area = (
        swat.normalize_thresholds(mesh, voronoi,
                                  thresh_dist=20.0,
                                  thresh_ridge=1.5,
                                  thresh_area=50.0,
                                  side=side))
    if path_to_mask:
        mask = io.load_texture(path_to_mask).darray[0]
    else:
        mask = None

    basins, ridges, adjacency = (
        swat.watershed(mesh,
                       voronoi,
                       dpf,
                       thresh_dist,
                       thresh_ridge,
                       thresh_area, mask))
    g = get_sulcal_graph(adjacency,
                         basins,
                         ridges,
                         save=True,
                         outdir=path_to_output)
    get_textures_from_graph(g, mesh, save=True, outdir=path_to_output)
    return g


def get_sulcal_graph(adjacency, basins, ridges, save=True, outdir=None):
    """
    Function that creates a graph from the outputs of the watershed.

    Node attributes are:
    - pit_index: index of the pit
    - pit_depth: depth of the pit
    - basin_vertices: list of vertices in the basin
    - basin_area: area of the basin
    - basin_label: label of the basin

    Edge attributes are:
    - ridge_index: index of the ridge
    - ridge_depth: depth of the ridge point
    - ridge_length: number of vertices in the ridge

    Parameters
    ----------
    adjacency
    basins
    ridges
    save
    outdir

    Returns
    -------

    """

    ##############################################################
    # Initialize the graph using adjacency matrix
    ##############################################################

    # As the adjacency matrix concerns all created basins during watershed,
    # it still contains merged basins that should not appear in the graph.
    # Thus, we first remove rows and columns corresponding to unconnected
    # basins (only zeros inside)
    labels = list(basins.keys())
    adjacency = adjacency[labels, :][:, labels]
    # np.fill_diagonal(adjacency, 1.)  # ensure systematic self connection
    graph = nx.from_numpy_array(adjacency)  # , nodelist=basins)
    # nodelist not adapted to attribution of labels in plotly_visu.py

    ####################################################################
    # Set graph attributes
    ####################################################################

    # Add node attributes
    node_attributes = {}
    for i, (label, values) in enumerate(basins.items()):
        node_attributes[i] = basins[label]  # add all dictionary values
        node_attributes[i]['basin_label'] = label  # add label value
    nx.set_node_attributes(graph, node_attributes)

    # Add edge attributes
    edge_attributes = {}
    for pair, values in ridges.items():
        # Get new indices
        i = labels.index(pair[0])
        j = labels.index(pair[1])
        edge_attributes[i, j] = values  # add all dictionary values
        print(values)
    for ed in range(len(graph.edges)):
        print("Edge "+str(ed)+" attributes:",
              graph.edges[list(graph.edges)[ed]].keys())

    nx.set_edge_attributes(graph, edge_attributes)
    for ed in range(len(graph.edges)):
        print("Edge "+str(ed)+" attributes:",
              graph.edges[list(graph.edges)[ed]].keys())

    if save:
        if not outdir:
            outdir = ''
        save_graph(graph, outdir)

    return graph


def save_graph(graph, outdir):
    """
    Save sulcal pits graph in the given directory under the
    name "graph.gpickle"
    Parameters
    ----------
    graph
    outdir

    Returns
    -------

    """

    file_path = os.path.join(outdir, "graph.gpickle")
    with open(file_path, 'wb') as f:
        pickle.dump(graph, f, pickle.HIGHEST_PROTOCOL)
    print("Graph saved in", file_path)
    return 0


def add_node_attribute_to_graph(graph, texture, name, save=True, outdir=None):
    """
    Add a node attribute to the graph using the value of the texture
    at pit positions
    Parameters
    ----------
    graph
    texture
    name
    save
    outdir

    Returns
    -------

    """

    if save and not outdir:
        outdir = ''

    node_values = {}
    for basin in graph.nodes:
        # Get pits indices
        pit = graph.nodes[basin]['pit_index']
        # Get the texture values for each pit
        node_values[basin] = texture[pit]

    # Add the attribute to the graph
    nx.set_node_attributes(graph, values=node_values, name=name)

    if save:
        save_graph(graph, outdir)

    return graph


def add_edge_attribute_to_graph(graph, texture, name, save=True, outdir=None):
    """
        Add an edge attribute to the graph using the value of the texture at
    ridge positions
    Parameters
    ----------
    graph
    texture
    name
    save
    outdir

    Returns
    -------

    """

    if save and not outdir:
        outdir = ''

    # Get the adjacency matrix with ridge positions
    adjacency = nx.to_numpy_array(graph, weight='ridge_index', dtype=np.int32)
    # Create and fill a new edge dictionary with the texture
    # values at ridge positions
    ridge_dict = {}
    for i, j in graph.edges:
        ridge_dict[(i, j)] = float(texture[adjacency[i][j]])
    # Add the attribute to the graph
    nx.set_edge_attributes(graph, ridge_dict, name=name)

    if save:
        save_graph(graph, outdir)

    return graph


def add_geodesic_distances_to_graph(graph, mesh, save=True, outdir=None):
    """
        Add the geodesic distances between ridge and pits to the corresponding
     ridge attributes in the graph:

    - geodesic_distance_btw_ridge_pit_i: geodesic distance between the
    ridge and the first pit
    - geodesic_distance_btw_ridge_pit_j: geodesic distance between the
    ridge and the second pit
    - geodesic_distance_btw_pits: geodesic distance between the two pits
    connected by the ridge (sum of previous values)
    Parameters
    ----------
    graph
    mesh
    save
    outdir

    Returns
    -------

    """
    if save and not outdir:
        outdir = ''

    # Create and fill a new edge dictionary with the geodesic distances
    geodistances = {}
    for i, j in graph.edges:
        ridge = graph.edges[(i, j)]['ridge_index']
        pit_i = graph.nodes[i]['pit_index']
        pit_j = graph.nodes[j]['pit_index']
        # Compute the geodesic distances between ridge and both pits
        gd_from_ridge = geodesics.compute_gdist(mesh, ridge)
        geodistances[(i, j)] = {}
        geodistances[(i, j)]['geodesic_distance_btw_ridge_pit_i'] = (
            float(gd_from_ridge[pit_i]))
        geodistances[(i, j)]['geodesic_distance_btw_ridge_pit_j'] = (
            float(gd_from_ridge[pit_j]))
        geodistances[(i, j)]['geodesic_distance_btw_pits'] = (
                float(gd_from_ridge[pit_i]) + float(gd_from_ridge[pit_j]))

    # Add the attribute to the graph
    nx.set_edge_attributes(graph, geodistances)

    if save:
        save_graph(graph, outdir)

    return graph


def add_mean_value_to_graph(graph, texture, name, save=True, outdir=None):
    """
        Add the mean value of the texture over the vertices of each basin
    to the graph node attributes
    Parameters
    ----------
    graph
    texture
    name
    save
    outdir

    Returns
    -------

    """
    if save and not outdir:
        outdir = ''

    average_values = {}
    for basin in graph.nodes:
        # Get the list of vertices
        vertices = graph.nodes[basin]['basin_vertices']
        # Compute the average value of texture over the vertices
        mean_value = np.mean(texture[vertices])
        average_values[basin] = mean_value

    # Add the attribute to the graph
    nx.set_node_attributes(graph, average_values, name=name)

    if save:
        save_graph(graph, outdir)

    return graph


def get_textures_from_graph(graph, mesh, save=True, outdir=None):
    """
    Function that returns the textures from a graph of sulcal pits
    Parameters
    ----------
    graph
    mesh
    save
    outdir

    Returns
    -------

    """
    if save and not outdir:
        outdir = ''

    vert = np.array(mesh.vertices)

    # texture of labels
    labels = np.full(len(vert), -1, dtype=np.int64)
    for b in graph.nodes:
        labels[graph.nodes[b]['basin_vertices']] = (
            graph.nodes)[b]['basin_label']
    tex_labels = texture.TextureND(darray=labels.flatten())
    if save:
        io.write_texture(tex_labels, os.path.join(outdir, "labels.gii"))

    # texture of pits
    atex_pits = np.zeros((len(vert), 1))
    pits_indices = list(nx.get_node_attributes(graph, 'pit_index').values())
    atex_pits[pits_indices] = 1
    tex_pits = texture.TextureND(darray=atex_pits.flatten())
    if save:
        io.write_texture(tex_pits, os.path.join(outdir,
                                                "pits_tex_from_graph.gii"))

    # texture of ridges
    atex_ridges = np.zeros((len(vert), 1))
    ridges_indices = (
        list(nx.get_edge_attributes(graph, 'ridge_index').values()))
    atex_ridges[ridges_indices] = 1
    tex_ridges = texture.TextureND(darray=atex_ridges.flatten())
    if save:
        io.write_texture(tex_ridges,
                         os.path.join(outdir, "rigdes_tex_from_graph.gii"))

    return tex_labels, tex_pits, tex_ridges



import unittest
import numpy as np
import slam.watershed as swat
import trimesh


def make_sphere():
    """Create a sphere"""
    mesh_a = trimesh.creation.icosphere(subdivisions=1, radius=1.0)
    return mesh_a


class TestWatershed(unittest.TestCase):

    def setUp(self):

        # Create a sphere
        self.sphere = make_sphere()

        # Create a simpler synthetic mesh
        self.mesh = self._create_mock_mesh()
        self.dpf = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
        self.voronoi = np.array([1, 1, 1, 1, 1])
        self.vert_neigh = {0: [1, 2], 1: [0, 3], 2: [0, 4], 3: [1], 4: [2]}
        self.thresh_dist = 0.3
        self.thresh_area = 2
        self.thresh_ridge = 1

    def _create_mock_mesh(self):
        class MockMesh:
            vertices = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0.5, 0.5, 0]])

        return MockMesh()

    def test_watershed(self):
        _, dpf, voronoi = swat.compute_mesh_features(self.sphere, save=False, outdir=None, check_if_exist=False)

        basins, ridges, adjacency = swat.watershed(
            self.sphere, voronoi, dpf, self.thresh_dist, self.thresh_ridge, self.thresh_area
        )

        self.assertIsInstance(basins, dict)
        self.assertIsInstance(ridges, dict)
        self.assertIsInstance(adjacency, np.ndarray)
        self.assertGreater(len(basins), 0, "At least one basin should be detected.")
        self.assertEqual(adjacency.shape[0], adjacency.shape[1], "Adjacency matrix should be square.")
        self.assertTrue(np.array_equal(adjacency, adjacency.T), "Adjacency matrix should be symmetric.")
        basins_with_ridge = []
        [[basins_with_ridge.append(b) for b in ridge_tuple] for ridge_tuple in ridges]
        self.assertTrue((np.sort(np.unique(basins_with_ridge)) == list(basins.keys())).all(), "Ridges and basins "
                                                                                              "dictionaries have "
                                                                                              "different basins labels")

    def test_get_textures_from_dict(self):
        basins = {
            0: {"basin_vertices": [0, 1], "pit_index": 0},
            1: {"basin_vertices": [2, 3], "pit_index": 2},
        }
        ridges = {
            (0, 1): {"ridge_index": 4, "ridge_depth": 0.5, "ridge_length": 1}
        }

        tex_labels, tex_pits, tex_ridges = swat.get_textures_from_dict(
            self.mesh, basins, ridges, save=False
        )

        self.assertIsNotNone(tex_labels)
        self.assertIsNotNone(tex_pits)
        self.assertIsNotNone(tex_ridges)
        self.assertEqual(tex_labels.darray.shape[1], len(self.mesh.vertices))
        self.assertEqual(tex_pits.darray.shape[1], len(self.mesh.vertices))
        self.assertEqual(tex_ridges.darray.shape[1], len(self.mesh.vertices))


if __name__ == '__main__':
    unittest.main()